#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <math.h>
#include <limits.h>
#ifdef IRIX
#  include <invent.h>
#endif
#ifdef Linux
#  include <values.h>
#endif

#include <Performer/pr.h>
#include <Performer/pfdu.h>
#include <Performer/pf/pfNode.h>
#include <Performer/pf/pfTraverser.h>
#include <dtkX11.h>

#include <dpf.h>
#include "IntersectionNode.h" 

dpfIntersectionNode::dpfIntersectionNode(void) {
  root = NULL;
  nHits = 0;

  //  Create the segment set in Performer shared memory.  Since this is
  //  a struct, we need to use pfMalloc instead of new
  void* arena = pfGetSharedArena();
  segset = (pfSegSet*) pfMalloc(sizeof(pfSegSet), arena);
  segset->mode = PFTRAV_IS_CULL_BACK | PFTRAV_IS_PRIM | PFTRAV_IS_PATH | PFTRAV_IS_GSET | PFTRAV_IS_GEODE;
  segset->activeMask = 0 ;
  segset->isectMask = -1;

}

//  Set the root of the scenegraph to do check for intersections
//  OK if rootNode == NULL
void dpfIntersectionNode::setRoot(pfNode* rootNode) {
  root = rootNode;
}

int dpfIntersectionNode::enableSegment(int segId) {
  //  Set bit segId in the active mask to 1
  if (segId >= PFIS_MAX_SEGS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::enableSegment- segId value greater",
	       "than allowed, %d>=%d.\n", segId,
	        PFIS_MAX_SEGS) ;
    return -1 ;
  }
  unsigned int mask = 1 << segId;
  segset->activeMask |= mask;  
  return 0 ;
}

int dpfIntersectionNode::disableSegment(int segId) {
  //  Set bit segId in the active mask to 0
  if (segId >= PFIS_MAX_SEGS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::disableSegment- segId value greater",
	       "than allowed, %d>=%d.\n", segId,
	        PFIS_MAX_SEGS) ;
    return -1 ;
  }
  int mask = ~(1 << segId);
  segset->activeMask &= mask;
  return 0 ;
}

//  Set the position of the given segment
int dpfIntersectionNode::setSegmentPos(int segId, float x, float y, float z) {
  if (segId >= PFIS_MAX_SEGS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::setSegmentPos- segId value greater",
	       "than allowed, %d>=%d.\n", segId,
	       PFIS_MAX_SEGS) ;
    return -1 ;
  }
  segset->segs[segId].pos.set(x, y, z);
  return 0 ;
}

//  Set the position of the given segment, using a vector
int dpfIntersectionNode::setSegmentPos(int segId, pfVec3 position) {
  return setSegmentPos(segId, position[0], position[1], position[2]);  
}

//  Set the direction of the given segment
int dpfIntersectionNode::setSegmentDir(int segId, float x, float y, float z) {
  if (segId >= PFIS_MAX_SEGS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::setSegmentDir- segId value greater",
	       "than allowed, %d>=%d.\n", segId,
	       PFIS_MAX_SEGS) ;
    return -1 ;
  }
  //  Normalize the vector
  float dist = sqrt(x*x + y*y + z*z);  
  if (dist == 0.f) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::setSegmentDir- (x,y,z) is zero.") ;
    return -1 ;
  }
  segset->segs[segId].dir.set(x/dist, y/dist, z/dist);
  return 0 ;
}

//  Set the direction of the given segment, using a vector
int dpfIntersectionNode::setSegmentDir(int segId, pfVec3 direction) {
  return setSegmentDir(segId, direction[0], direction[1], direction[2]); 
} 

//  Set the length of the specified segment
int dpfIntersectionNode::setSegmentLength(int segId, float len) {
  if (segId >= PFIS_MAX_SEGS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::setSegmentLength- segId value greater",
	       "than allowed, %d>=%d.\n", segId,
	       PFIS_MAX_SEGS) ;
    return -1 ;
  }
  segset->segs[segId].length = len;  
  return 0 ;
}

//  Check for intersections with the current segment set
void dpfIntersectionNode::doIntersection() {
  if (root) 
    root->isect(segset, hits);
}

//  Check to see if the specified segment recorded a hit
//  Returns -1 on error, 0 if no hit, 1 if a hit
int dpfIntersectionNode::segmentHit(int segId) {
  if (segId >= PFIS_MAX_SEGS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::setSegmentHit- segId value greater",
	       "than allowed, %d>=%d.\n", segId,
	       PFIS_MAX_SEGS) ;
    return -1 ;
  }

  int flags;
  hits[segId][0]->query(PFQHIT_FLAGS, &flags);
  return (flags & PFHIT_ISECT);
}

//  Get the local (non-transformed) point of intersection for the specified
//  segment.
int dpfIntersectionNode::getLocalPoint(int segId, pfVec3* point) {
  if (segId >= PFIS_MAX_SEGS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::getLocalPoint- segId value greater",
	       "than allowed, %d>=%d.\n", segId,
	       PFIS_MAX_SEGS) ;
    return -1 ;
  }
  
  if (!segmentHit(segId)) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::getLocalPoint- ",
	       "requesting intersection point when the segment ",
	       "did not intersect\n");
    return -1 ;
  }
  hits[segId][0]->query(PFQHIT_POINT, point);
  return 0 ;
}

// Get the path of the node that is hit by the specified segment
//  Retuns 0 if successful
pfPath* dpfIntersectionNode::getHitPath(int segId) {
  pfPath *hitPath ;
  if (segId >= PFIS_MAX_SEGS) {
    printf("%s, %d\n",__FILE__,__LINE__) ;
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::getHitPath- segId value greater"
	       "than allowed, %d>=%d.\n", segId,
	       PFIS_MAX_SEGS) ;
    return NULL ;
  }
  
  if (!segmentHit(segId)) {
    printf("%s, %d\n",__FILE__,__LINE__) ;
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::getHitPath- "
	       "requesting intersection point when the segment "
	       "did not intersect\n");
    return NULL ;
  }
  hits[segId][0]->query(PFQHIT_PATH, &hitPath);
  // printf("%s, %d\n",__FILE__,__LINE__) ;
  // printf("num = %d, len = %d\n",hitPath->getNum(),hitPath->getArrayLen()) ;
  return hitPath ;
}

//  Get the point of intersection for the specified segment, transformed
//  to the coordinate system of the root node
int dpfIntersectionNode::getXFormPoint(int segId, pfVec3* point) {
  
  if (segId >= PFIS_MAX_SEGS) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::getXformPoint- segId value greater",
	       "than allowed, %d>=%d.\n", segId,
	       PFIS_MAX_SEGS) ;
    return -1 ;
  }
  
  pfVec3 locPoint;
  getLocalPoint(segId, &locPoint);
  
  pfMatrix xform;
  hits[segId][0]->query(PFQHIT_XFORM, &xform);
  point->xformPt(locPoint, xform);
  return 0 ;
  
}

dpfIntersectionNode::~dpfIntersectionNode() {
  if (segset)
    pfDelete(segset);
  segset = NULL ;
}

//  Convenience method to make an object invisible to Intersection detection
int dpfIntersectionNode::noIntersection(pfNode* node) {
  if (!node) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::noIntersection- node is NULL.\n") ;
    return -1 ;
  }
  node->setTravMask(PFTRAV_ISECT, 0, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);
  return 0 ;
}

//  Convenience method to make an object visible again
int dpfIntersectionNode::yesIntersection(pfNode* node) {
  if (!node) {
    dtkMsg.add(DTKMSG_ERROR,
	       "dpfIntersectionNode::noIntersection- node is NULL.\n") ;
    return -1 ;
  }
  node->setTravMask(PFTRAV_ISECT, 1, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);
  return 0 ;
}

