// OMS - This is a DIVERSE DSO which loads files specified by the
// environment variable DPF_OMS_FILES. Each of these files has a
// set of lines specifying a location (xyz hpr) and filename for an object.
// The program reads in all the lines from all the files and displays
// each object according to its xyz hpr position. Objects can be
// moved around using the wand and its buttonys. An object can be
// selected by the wand by pressing button 1. The object then moves
// with the wand by holding button 1. Switch to button 0 to stop
// the object and select a new object.
//
// Authors:
//     Adele Peskin
//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/
//  
//    Steve Satterfield (SAVG) provided the idea for the program
//
//
// For information about this DSO, 
// contact Adele Peskin, peskin@boulder.nist.gov
//
//
// This code is a modification to the example generic dtkAugment DSO provided
// with the DIVERSE distribution. The following disclaimer applies to the NIST
// modifications.
// 
// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  OMS is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <fstream.h>
#include <dtk.h>
#include "IntersectionNode.C"

// streams
ifstream data_file;

// example Oms dtkAugment DSO
class Oms : public dtkAugment {
public:
  Oms(dpf *app) ;
  ~Oms(void) ;
  int preFrame(void);
  int postConfig(void);
  int postFrame(void);
  
private:
  dpf* app;
  dpfDisplay *display;
  //  Shared memory segment for the wand data
  dtkInLocator *wand ;
  float wandData[6];
  float oldWandData[6];
  pfCoord wandCoord;

  //  Shared memory segment for the button that switches rotation methods
  dtkSharedMem* shm;
  //  Shared memory segment used to request the wand pointer from another program
  dtkSharedMem* shm2;
  //  Shared memory segment used to request a sound from another program
  dtkSharedMem* shm3;
  signed char makeSound;
  signed char selected;
  pfVec3 check;
  pfVec3 check2;
  pfVec3 oldCheck;
  float firstDistance;
  float wandDistance;
  dtkInButton *button;
  dtkDequeuer *dequeuer;
  // counter for the dcs's
  int cnt;
  // counter for the object files
  int filecnt;
  pfDCS** dcsList;
  pfGroup* rootGroup;
  pfNode** objectPtrs;
  char** file_names;
  int* fileNumbers;
  int selected_file_number;
  bool objectSelected;
  bool button1On; // indicates if button 1 is being pressed
  bool button1LastFrame; // indicates old position of button 1
  int last_one_hit;
  int frame_timer;
  int oldButton;
  int justSelected;
  ofstream movingdata;
  float *xpt, *ypt, *zpt, *hpt, *ppt, *rpt;

  dpfIntersectionNode* intersect;
  bool testNames(char* name1, char* name2);
  int testSuffix(char* nameStore);
  pfDCS* findDCS(pfPath* path);
  float findDistance(pfVec3 ptA, float* ptB);
  void moveCenterToWandLine(void);
  void keepCenterOnWandLine();
  void findNewPoint();
  void writeOutput(void);
  pfDCS* currentDCS;
  pfVec3 intPoint;
  float delta[6];
  bool dontMoveFirst;
  pfMatrix* allMove;
  int rotation;
} ;

// create object and set description.
// be sure to validate it!
Oms::Oms(dpf *app_in) :
  dtkAugment("Oms")
{
  setDescription(" loads files specified by DPF_OMS_FILES"
		 " environment variable and displays an object from"
		 " each file at a xyzhpr position also specified"
                 " in the file. The user uses the wand to place"
                 " the objects in a new location.") ;
  cout << "objectMover version compiled on 3/03/03" << endl;
  app = app_in;
  validate();
  cnt = 0 ;
  filecnt = 0;
  dcsList = NULL;
  objectPtrs = NULL;
  file_names = NULL;
  fileNumbers = NULL;
  objectSelected = false;
  frame_timer = 0;
  //  Check if there is a segment "ABswitch" to control the ABswitch rate
  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  if(!wand || wand->isInvalid())
    wand = NULL ;
  button = NULL;
  dequeuer = NULL;
  // check to see if there is a segment "rotate" to change the mode of rotation
  shm = new dtkSharedMem(sizeof(int), "rotateInt");
  //int switchRotation = 0;
  if(!shm || shm->isInvalid()) {
    shm = NULL ;
  }
  rotation = 0;

  shm2 = new dtkSharedMem(sizeof(unsigned char), "pointerOn");
  if(!shm2 || shm2->isInvalid()) {
      shm2 = NULL;
  }

  //  Check if there is a segment "makeSound" to start the sound
  shm3 = new dtkSharedMem(sizeof(unsigned char), "makeSound");
  if(!shm3 || shm3->isInvalid()) {
      shm3 = NULL;
  }
  
  // turn buttons off to start for makeSound and buttonval
  unsigned char turnit = 0;
  if (shm3)
     shm3->write(&turnit);

  last_one_hit = -1;
  firstDistance = -1.0;
  wandDistance = -1.0;
  currentDCS = NULL;
  oldButton = 0;
  dontMoveFirst = true;
  button1On = false;
  button1LastFrame = false;
  // justSelected: flag to check at button change
  // only write to file if button changed from 1 to 0 
  // AND an object was selected
  justSelected = 0;
  selected = 2;
  xpt = NULL;
} ;

// destructor
Oms::~Oms(void) {
  writeOutput( );
  movingdata.close();
}

#define MIN_BUTTONS 3

int Oms::preFrame(void)
{
  dequeuer = new dtkDequeuer(app->record());
  if(!dequeuer || dequeuer->isInvalid()) return DTK_CALLBACK_ERROR;

  display = app->display();
  if(!display) return DTK_CALLBACK_ERROR;

  button = (dtkInButton *) app->get("buttons", DTKINBUTTON_TYPE);
  if(!button) return DTK_CALLBACK_ERROR;

   if(button->number() < MIN_BUTTONS)
     {
       dtkMsg.add(DTKMSG_ERROR,
		  "buttonNavControl::preConfig() failed:"
		  " the dtkInButton named \"%s\" has "
		  "just %d buttons"
		  " which is less than the %d needed.\n",
		  button->getName(), button->number(),
		  MIN_BUTTONS);
      return DTK_CALLBACK_ERROR;
    }
  button->queue();

  // just call buttonNavControl::preFrame() once at the start.
  return DTK_REMOVE_CALLBACK;
}

// called only once by dpf, after pfConfig() and dpfDisplay:postConfig()
// pfConfig() 
int Oms::postConfig(void) {

  char *envInput = getenv("DPF_OMS_FILES") ;
  char *envOutput = getenv("DPF_OUT_FILE") ;
  char *original_list ;
  char* file_a = new char[1000];
  xpt = new float[1000];
  ypt = new float[1000];
  zpt = new float[1000];
  hpt = new float[1000];
  ppt = new float[1000];
  rpt = new float[1000];
  int spaceHPR = 1000;
  
  // as the files are read, a dcs is formed and added to a group
  // node; a matrix from the xyz hpr values is set in the dcs

  // each object and its corresponding matrix as a DCS
  // into a group node
 
  // first get the display and make and set the group node
  display = app->display();
  rootGroup = new pfGroup;
  display->world()->addChild(rootGroup);

  pfDCS* dcs ;
  pfNode* newObject;
  
  // if there are environment variables for input and output, use them;
  // otherwise, use a default
  if (!envInput) {
    envInput = "xyzList";
  }
  if (!envOutput) {
    envOutput = "movingPos";
  }
  // open the output file for the object locations
  movingdata.open ( envOutput, ios::out );
  movingdata.precision(12);

  // see if there is an environment variable for the output file
  // read in the list of file names
  original_list = strdup(envInput) ;
  // strtok() is broken in GNU libc, so we roll our own here.
  // copy the original list
  char *end = original_list ;
  for(;*end;) {
    for(; *end != ':' && *end; end++);
    if(*end) { *end = '\0'; end++; }
    // add to the list of files whose names are in original list
    if (strlen(original_list)>0) {
      // read each file
      data_file.open(original_list, ios::in);
      printf("next file off of the list \"%s\"\n",original_list) ;
      while (data_file >> xpt[cnt] >> ypt[cnt] >> zpt[cnt] >> hpt[cnt] >> ppt[cnt] >> rpt[cnt] >> file_a) {
	// create a new dcs for the new xyzhpr and the newObject node
	dcs = new pfDCS();
	rootGroup->addChild(dcs);
	
	// see if this object is new
	bool isItNew = true;
	int place = -1;
	for ( int i = 0; i < filecnt; i++ ) {
	  if (testNames(file_a,file_names[i])) {
	    isItNew = false;
	    place = i;
	  }
	}
	
	// make room for another fileNumbers entry
	fileNumbers = (int *) realloc(fileNumbers, sizeof(int)*(cnt+1));
	
	if (isItNew) {
	  // load the file into a pfNode
	  newObject = pfdLoadFile(file_a);
	  file_names = (char **) realloc(file_names, sizeof(char*)*(filecnt+1));
	  // make space for the new file name in file_names
	  file_names[filecnt] = file_a;
	  objectPtrs = (pfNode **) realloc(objectPtrs, sizeof(pfNode*)*(filecnt+1));
	  objectPtrs[filecnt] = newObject;
	  
	  // list this file as the one for the current object
	  fileNumbers[cnt] = filecnt;
	  ++filecnt;
	}
	else {
	  //pull out the old node pointer
	  newObject = objectPtrs[place];
	  // list this file as the one for the current object
	  fileNumbers[cnt] = place;
	} 
	
	// load the dcs with the object info
	dcs->addChild(newObject) ;
	dcs->setRot(hpt[cnt], ppt[cnt], rpt[cnt]);
	dcs->setTrans(xpt[cnt], ypt[cnt], zpt[cnt]);
	
	// add the dcs to the dcs list, newObject to the pfnode list,
	// the name of the file to the file name list
	dcsList = (pfDCS **) realloc(dcsList, sizeof(pfDCS*)*(cnt+1));
	dcsList[cnt] = dcs;
	
	// reallocate space to store the translation values
	cnt++ ;
	
	// load the converter if necessary
	char *tempa = file_a;
	int err = testSuffix(tempa);
	original_list = end;
	// reassign new space for file_a
	file_a = new char[1000];

	// see if more space for hpt,ppt,rpt is needed
        if ((cnt+1) == spaceHPR) {
	  spaceHPR += 1000;
	  xpt = (float *) realloc(xpt, sizeof(float)*(spaceHPR));
	  ypt = (float *) realloc(ypt, sizeof(float)*(spaceHPR));
	  zpt = (float *) realloc(zpt, sizeof(float)*(spaceHPR));
	  hpt = (float *) realloc(hpt, sizeof(float)*(spaceHPR));
	  ppt = (float *) realloc(ppt, sizeof(float)*(spaceHPR));
	  rpt = (float *) realloc(rpt, sizeof(float)*(spaceHPR));
	}
      } //end while
      data_file.close();
    }
  }

  //  Initialize the intersection detection.  Segment 0 will contain a line
  //  segment starting at the wand position and extending 100 units in the 
  //  direction of the wand.  This will be in scene coordinates.
  intersect = new dpfIntersectionNode();
  intersect->enableSegment(0);
  intersect->setRoot(rootGroup);

  // get the scale to set the segment length
  float lengthScale = display->scale();
  if (abs(lengthScale) > 1.0e-6)
   intersect->setSegmentLength(0, 100.0/lengthScale);
  else
   intersect->setSegmentLength(0, 1.0);

  // keep the matrices of the objects' dcs's for movement transformations
  allMove = new pfMatrix[cnt];

  // set the matrices
  for ( int i = 0; i < cnt; ++i ) {
    allMove[i].makeTrans(xpt[cnt],ypt[cnt],zpt[cnt]);
    allMove[i].makeEuler(hpt[cnt],ppt[cnt],rpt[cnt]);
  }
  return DTK_REMOVE_CALLBACK ; 
}


bool Oms::testNames(char* name1, char* name2) {
   bool sameName = true;
   int next = 0;
   for ( int j = 0; j < 10; ++j ) {
     if (name1[next] != name2[next]) {
       sameName = false;
       return sameName;
     }
     ++next;
   }
   return sameName;
}

int Oms::testSuffix(char* nameStore) {
  int sufc = 0;
  char *suf ;
  char **sufv = NULL ;
  if (strlen(nameStore)>0) {
    suf = strrchr(nameStore, '.') ;
    if (suf) {
      int suf_found = 0 ;
      for (int c = 0; c<sufc; c++) {
	if (!strcmp(suf,sufv[c])) {
	  suf_found = 1 ;
	  break ;
		    
	}
      }
      if (!suf_found) {
	sufv = (char **) realloc(sufv, sizeof(char *)*(sufc+1));
	sufv[sufc] = strdup(suf) ;
	sufc++ ;
	if(!pfdInitConverter(suf)) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "ABCDswitch::preConfig: pfdInitConverter(%s) failed.\n",
		     suf) ;
	  return DTK_CALLBACK_ERROR;
	}
      }
    } else {
      dtkMsg.add(DTKMSG_ERROR,
		 "ABCDswitch::preConfig: no file suffix found for file \"%s\"\n",nameStore) ;
      return DTK_CALLBACK_ERROR;
    }
	      
    if(!pfdInitConverter(suf)) {
      dtkMsg.add(DTKMSG_ERROR,
		 "pfdInitConverter(%s) failed.\n",
		 suf) ;
      return DTK_CALLBACK_ERROR;
    }      
  }
  return 0;
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int Oms::postFrame(void) {

  // first read the button (whichbutton) on the mouse/wand
  int whichButton;
  dtkRecord_event *e = dequeuer->getNextEvent(button);
  int buttonState = button->read(&whichButton, e);
  int i;


  //--------------------------------------------------
  // button1On; // switches back and forth when button 1 is pressed
  // button1LastFrame; // indicates old position of button 1
  //--------------------------------------------------

  if(whichButton == 1) {
    // first see if button 1 is currently pressed
    if (button->read(1,e)) {
      if (!button1LastFrame) {
	// switch button1ON: if off, turn it on; if on, turn it off
	if (button1On) 
	  button1On = false;
	else
	  button1On = true;
	//if (button1On) cout << "object will move now" << endl;
	//if (!button1On) cout << "object will stop moving" << endl;
      }
      button1LastFrame = true;
    }
    else
      button1LastFrame = false;
  }
  //--------------------------------------------------


  //if (whichButton == 1) {
  if (button1On) {

    // read the wand direction
    if (wand)
      wand->read(wandData);

    // read the rotation mode
    if (shm) {
      shm->read(&rotation);
    }

    // define the wand data in world coordinates
    wandCoord.xyz.set(wandData[0], wandData[1], wandData[2]);
    wandCoord.hpr.set(wandData[3], wandData[4], wandData[5]);
    display->dpfToWorld(&wandCoord);
    
    // create a matrix from the HPR position in world coordinates
    pfMatrix wandWorld;
    wandWorld.makeEuler(wandCoord.hpr[0],wandCoord.hpr[1],wandCoord.hpr[2]);

    // create a vector in the +y direction and move according to the wand
    pfVec3 upVector(0.f, 1.f, 0.f);
    check.xformVec(upVector, wandWorld);

    //  Set up segment 0 and do the intersection
    intersect->setSegmentPos(0, 
			     wandCoord.xyz[0], 
			     wandCoord.xyz[1], 
			     wandCoord.xyz[2]);
    intersect->setSegmentDir(0, check);
    intersect->doIntersection() ;
    int j = intersect->segmentHit(0);

    // if no object is selected, pick one
    if (!objectSelected) {

      if (j>0) {
	// get the path to the hit to find the dcs to update
	pfPath* path = intersect->getHitPath(0) ;
	if (path) {
	  // select the dcs to update
	  pfDCS* hitDCS = findDCS(path);
	  if (hitDCS != NULL ) {
	    // get the intersection point
	    intersect->getXFormPoint(0, &intPoint);
	    // this is the first time hit - compute the distance from
	    // the wand to the object
	    firstDistance = findDistance (intPoint, wandData);

	    objectSelected = true;
	    // make a sound to indicate that the object was selected
	    makeSound = 2;
	    if (shm3)
	      shm3->write(&makeSound);

	    frame_timer = 1;

	    // set this DCS to the current DCS - no other DCS can be
	    // picked until the button is released
	    currentDCS = hitDCS;

	    // find the center of this object and move the object
	    // so that the center is along the line with the wand
	    moveCenterToWandLine();
	  }
	}
      }
    }
    else { // object is already selected, move it unless button 1 repressed
      ++frame_timer;
      if (frame_timer > 0 ) {
	// every 10 frames,
	// place object at the defined distance where the wand is now
	if (button1On) {
	  findNewPoint();
	  // move the xyzhpr points unless this is the first move
	  if (!dontMoveFirst) {
	    keepCenterOnWandLine();
	    // remake the dcs matrix
	    allMove[last_one_hit].makeTrans(xpt[last_one_hit],
					    ypt[last_one_hit],
					    zpt[last_one_hit]);
	    allMove[last_one_hit].makeEuler(hpt[last_one_hit],
					    ppt[last_one_hit],
					    rpt[last_one_hit]);
	  }
	  else {
	    dontMoveFirst = false;
	  }
	  frame_timer = 0;
	  objectSelected = true;
	  selected_file_number = last_one_hit;
	  justSelected = 1;
	  // moved the object, so save the wandData as oldWandData
	  for ( i = 0; i < 3; i++ ) {
	    oldWandData[i] = wandCoord.xyz[i];
	  }
	  for ( i = 3; i < 6; i++ ) {
	    oldWandData[i] = wandCoord.hpr[i-3];
	  }
	  for (i = 0; i < 3; i++ ) {
	    oldCheck[i] = check[i];
	  }
	}
	else {
	  // button1Off - restart
	  //writeOutput( );
	  oldButton = 2;
	  justSelected = 0;
	  objectSelected = false;
	  // set this so that the first touch of the wand does not send an object
	  // where it shouldn't be - need one step to get a delta vector
	  dontMoveFirst = true;
	  for ( i = 0; i < 3; i++ ) {
	    oldWandData[i] = wandCoord.xyz[i];
	  }
	  for ( i = 3; i < 6; i++ ) {
	    oldWandData[i] = wandCoord.hpr[i-3];
	  }
	  // make a sound to indicate that the object was deselected
	  makeSound = 3;
	  if (shm3)
	    shm3->write(&makeSound);
	}
      }
    }
  }
  else { // no hit - zero out last_hit number and last position arrays
    last_one_hit = -1;
    frame_timer = 0;
    currentDCS = NULL;
    oldButton = 0;
    for (int i = 0; i < 3; i++ ) {
      oldCheck[i] = check[i];
    }
    // button1Off - restart
    //writeOutput( );
    oldButton = 2;
    justSelected = 0;
    // set this so that the first touch of the wand does not send an object
    // where it shouldn't be - need one step to get a delta vector
    dontMoveFirst = true;
    for ( i = 0; i < 6; i++ ) {
      oldWandData[i] = 0.0;
    }
    // make a sound to indicate that the object was deselected
    if (objectSelected) {
      makeSound = 3;
      if (shm3)
	shm3->write(&makeSound);
    }
    objectSelected = false;
  }

  // write whether the object is selected to the shared memory 
  if (shm2) {
    //if (!objectSelected && button1On )
    if (button1On )
      // show the wand when button 1 is pressed,
      // but the object is not yet located
      selected = 1;
    else
      // hide the wand
      selected = 2;
    shm2->write(&selected);
  }
  return DTK_CONTINUE ;
}

  pfDCS* Oms::findDCS(pfPath* path) {
  // loop through the cnt list of group nodes and find one on this path
  pfDCS* foundPtr = NULL;
  bool found = false;
  for (int i=0; i<path->getNum(); ++i) {
    if (found) break;
    for ( int j = 0; j < cnt; j++ ) {
      // look for dcs nodes in the path
      if ((path->get(i)) == dcsList[j]) {
	foundPtr = dcsList[j];
        // found the right one
	last_one_hit = j;
	found = true;
	break;
      }
    }
  }
  return foundPtr;
}

void Oms::moveCenterToWandLine(void){
  pfSphere* psphere = new pfSphere();
  currentDCS->getBound(psphere);
  pfVec3 centr = psphere->center;

  // the line from the wand is defined by a point(wandCoord) 
  // and a direction (check) r = r0 + d(direction);
  // so, need to find the distance d up the line that is perpendicular
  // to the center point, that is at the point P=(a,b,c)

  // the dot product of the P-centr with P-wand = 0 if the vectors are
  // perpendicular - did some algebra and came up with an expression for d:
  float d = check[0]*(centr[0]-wandCoord.xyz[0]) +
            check[1]*(centr[1]-wandCoord.xyz[1]) +
            check[2]*(centr[2]-wandCoord.xyz[2]);
  d = d/(check[0]*check[0] + check[1]*check[1] + check[2]*check[2]);

  // the point along the wind line is wandCoord.xyz + d *(check)
  // move the object's center to this point
  
  xpt[last_one_hit] += wandCoord.xyz[0] + d*check[0] - centr[0];
  ypt[last_one_hit] += wandCoord.xyz[1] + d*check[1] - centr[1];
  zpt[last_one_hit] += wandCoord.xyz[2] + d*check[2] - centr[2];
  currentDCS->setTrans(xpt[last_one_hit],
		       ypt[last_one_hit],
		       zpt[last_one_hit]);

  // save distance
  wandDistance = d;
}

void Oms::keepCenterOnWandLine() {
  // find the point the distance wandDistance down the wand
  
  xpt[last_one_hit] = wandCoord.xyz[0] + wandDistance*check[0];
  ypt[last_one_hit] = wandCoord.xyz[1] + wandDistance*check[1];
  zpt[last_one_hit] = wandCoord.xyz[2] + wandDistance*check[2];
  //cout << "new coords: " << xpt[last_one_hit]<<" "<<ypt[last_one_hit]<<" "<<zpt[last_one_hit]<<endl;
  currentDCS->setTrans(xpt[last_one_hit],
		       ypt[last_one_hit],
		       zpt[last_one_hit]);
  
}

float Oms::findDistance(pfVec3 ptA, float* ptB) {
  return sqrt((ptA[0]-ptB[0])*(ptA[0]-ptB[0]) +
              (ptA[1]-ptB[1])*(ptA[1]-ptB[1]) +
              (ptA[2]-ptB[2])*(ptA[2]-ptB[2]));
}

void Oms::findNewPoint() {
  if (!dontMoveFirst) {

    // find delta = translation of object
    delta[0] = wandCoord.xyz[0] - oldWandData[0];
    delta[1] = wandCoord.xyz[1] - oldWandData[1];
    delta[2] = wandCoord.xyz[2] - oldWandData[2];
    delta[3] = wandCoord.hpr[0] - oldWandData[3];
    delta[4] = wandCoord.hpr[1] - oldWandData[4];
    delta[5] = wandCoord.hpr[2] - oldWandData[5];

    // the translation can just happen, but the rotation needs
    // to be done in the object's coordinate system
    
    // multiply allMove by delta(345)
    float hh,pp,rr;
    pfVec3 row;
    allMove[last_one_hit].getRow(0, row);
    hh = row[0]*delta[3] +
         row[1]*delta[4] +
         row[2]*delta[5];
    allMove[last_one_hit].getRow(1, row);
    pp = row[0]*delta[3] +
         row[1]*delta[4] +
         row[2]*delta[5];
    allMove[last_one_hit].getRow(2, row);
    rr = row[0]*delta[3] +
         row[1]*delta[4] +
         row[2]*delta[5];
    
    if (rotation == 1) {
      delta[3] = hh;
      delta[4] = pp;
      delta[5] = rr;
    }
    else {
      // the default - the object should move exactly with the wand
      // convert the change in the rotation to a translational move
      delta[3] = 0.0;
      delta[4] = 0.0;
      delta[5] = 0.0;
      // first find the vector difference in rotational move
      for (int i = 0; i < 3; i++ ) {
	delta[i] += check[i] - oldCheck[i];
      }
    }
  }
  else {
    // start
    for ( int i = 0; i < 6; i++ ) {
      delta[i] = 0.0;
    }
  }
}

void Oms::writeOutput( ) {
  // write the current position of each object
  for ( int i = 0; i < cnt; i++ ) {
    // write the name of the object and its current position to output
    movingdata << xpt[i] << "  " <<  
                  ypt[i] << "  " << 
                  zpt[i] << "  " << 
                  hpt[i] << "  " <<  
                  ppt[i] << "  " <<  
                  rpt[i] << "  " << 
                  file_names[i] << endl;
  }
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new Oms(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
